#version 330 core
#extension GL_ARB_explicit_uniform_location : enable
#extension GL_ARB_separate_shader_objects : enable

in vec3 fragNormal;
in vec2 fragUV;
in vec3 fragPosition;

layout(location = 0) out vec4 outColor;
layout(location = 1) out vec3 outNormal;
layout(location = 2) out vec3 outDepth;

layout(location = 0) uniform mat4 model;
uniform vec3 cameraPosition;

layout(location = 2) uniform sampler2D tex;
uniform float materialShininess;
uniform vec3 materialSpecularColor;

uniform int numLights;
struct Light{
	vec3 position;
	vec3 color;
	float attenuation;
	float ambientCoefficient;
	float coneAngle;
	vec3 coneDirection;
	mat4 depthVP;
};
uniform Light lights[20];
uniform sampler2DShadow shadowMap0;
uniform sampler2DShadow shadowMap1;
uniform sampler2DShadow shadowMap2;
uniform sampler2DShadow shadowMap3;
uniform sampler2DShadow shadowMap4;
uniform sampler2DShadow shadowMap5;
uniform sampler2DShadow shadowMap6;
uniform sampler2DShadow shadowMap7;
uniform sampler2DShadow shadowMap8;
uniform bool doShadowing;

uniform bool doTextureIllumination;

float bias = 0.001;
vec2 poissonDisk[4] = vec2[] (
  vec2( -0.94201624, -0.39906216 ),
  vec2( 0.94558609, -0.76890725 ),
  vec2( -0.094184101, -0.92938870 ),
  vec2( 0.34495938, 0.29387760 )
);

vec3 cellShade(vec3 inColor, int numShades)
{
	numShades--;
	vec3 outColor;
	outColor.x = ceil(inColor.x * numShades)/numShades;
	outColor.y = ceil(inColor.y * numShades)/numShades;
	outColor.z = ceil(inColor.z * numShades)/numShades;
	outColor += vec3(0.3, 0.3, 0.3);
	outColor = pow(outColor, vec3(1.5, 1.5, 1.5));
	outColor -= vec3(0.3, 0.3, 0.3);
	return outColor;
}

vec3 calcLight(Light light, vec3 surfaceColor, vec3 normal, vec3 surfacePos, vec3 surfaceToCamera, sampler2DShadow shadowMap)
{
	vec3 surfaceToLight;
	float attenuation = 1.0;
	float visibility = 1.0;
	vec4 shadowCoords;

	// calc ambient
	vec3 ambient = light.ambientCoefficient * surfaceColor.rgb * light.color;

	// calc visibility
	if (doShadowing && light.attenuation == 0.0)
	{
		shadowCoords = light.depthVP * model * vec4(fragPosition, 1);
		for (int i = 0; i < 4; i++){
			visibility += texture(shadowMap, vec3(shadowCoords.xy + poissonDisk[i]/8000.0, shadowCoords.z/shadowCoords.w - bias));
		}
		visibility /= 4;
	}

	if (visibility > 0.0)
	{
		if(light.attenuation == 0.0)	// directional light
		{
			surfaceToLight = normalize(light.position);
		}
		else							// calc attenuation
		{
			surfaceToLight = normalize(light.position - surfacePos);
			vec3 coneDirection = normalize(light.coneDirection);
			vec3 rayDirection = -surfaceToLight;
			float lightToSurfaceAngle = degrees(acos(dot(rayDirection, coneDirection)));
			if (lightToSurfaceAngle > light.coneAngle)
			{
				attenuation = 0.0;
			}

			if (attenuation != 0.0)
			{
				float distanceToLight = length(light.position - surfacePos);
				attenuation = 1.0 / (1.0 + light.attenuation * pow(distanceToLight, 2));
			}
			
			if (attenuation < 0.01)
			{
				return ambient;		// drop out if to low
			}
		}
	}
	else
	{
		return ambient;		// drop out if not visible
	}
	
	// calc diffuse
	float diffuseCoefficient = max(0.0, dot(normal, surfaceToLight));
	vec3 diffuse = diffuseCoefficient * surfaceColor.rgb * light.color;

	// calc specular
	float specularCoefficient = 0.0;
	if(diffuseCoefficient > 0.0) specularCoefficient = pow(max(0.0, dot(surfaceToCamera, reflect(-surfaceToLight, normal))), 1/materialShininess);
	vec3 specular = specularCoefficient * materialSpecularColor * light.color;

	return ambient + visibility * attenuation * (diffuse + specular);	
}

void main()
{
	vec3 normal = normalize(transpose(inverse(mat3(model))) * fragNormal);
	outNormal = normal;
	float depthVal = gl_FragCoord.z;
	depthVal -= 0.985;
	depthVal *= 62.5;
	outDepth = vec3(depthVal, depthVal, depthVal);
	vec3 surfacePos = vec3(model * vec4(fragPosition, 1));
	vec4 surfaceColor = texture(tex, fragUV);
	vec3 surfaceToCamera = normalize(cameraPosition - surfacePos);

	if ( doTextureIllumination &&
		(surfaceColor.x > 0.90 || surfaceColor.x < 0.10) && 
		(surfaceColor.y > 0.90 || surfaceColor.y < 0.10) &&
		(surfaceColor.z > 0.90 || surfaceColor.z < 0.10) &&
		!(surfaceColor.x < 0.10 && surfaceColor.y < 0.10 && surfaceColor.z < 0.10)
	)
	{
		outColor = surfaceColor;
	}
	else
	{
		surfaceColor = vec4(cellShade(surfaceColor.xyz, 5), surfaceColor.w);

		vec3 linearColor = vec3(0);
		for(int i = 0; i < numLights; ++i)
		{
			if(distance(fragPosition, lights[i].position) < 40 || lights[i].attenuation == 0.0)
			{
				switch(i){
					case 0:
						linearColor += calcLight(lights[i], surfaceColor.rgb, normal, surfacePos, surfaceToCamera, shadowMap0);
						break;
					case 1:
						linearColor += calcLight(lights[i], surfaceColor.rgb, normal, surfacePos, surfaceToCamera, shadowMap1);
						break;
					case 2:
						linearColor += calcLight(lights[i], surfaceColor.rgb, normal, surfacePos, surfaceToCamera, shadowMap2);
						break;
					case 3:
						linearColor += calcLight(lights[i], surfaceColor.rgb, normal, surfacePos, surfaceToCamera, shadowMap3);
						break;
					case 4:
						linearColor += calcLight(lights[i], surfaceColor.rgb, normal, surfacePos, surfaceToCamera, shadowMap4);
						break;
					case 5:
						linearColor += calcLight(lights[i], surfaceColor.rgb, normal, surfacePos, surfaceToCamera, shadowMap5);
						break;
					case 6:
						linearColor += calcLight(lights[i], surfaceColor.rgb, normal, surfacePos, surfaceToCamera, shadowMap6);
						break;
					case 7:
						linearColor += calcLight(lights[i], surfaceColor.rgb, normal, surfacePos, surfaceToCamera, shadowMap7);
						break;
					case 8:
						linearColor += calcLight(lights[i], surfaceColor.rgb, normal, surfacePos, surfaceToCamera, shadowMap8);
						break;
					default:
						break;
				}
			}
		}

		vec3 gamma = vec3(1.0/2.2);
		outColor = vec4(pow(linearColor, gamma), surfaceColor.a);
	}
	
}