#version 330 core
#extension GL_ARB_explicit_uniform_location : enable
#extension GL_ARB_separate_shader_objects : enable

layout(points) in; 
layout(points) out; 
layout(max_vertices = 8) out; 

// All that we get from vertex shader

in vec3 positionPass[]; 
in vec3 velocityPass[]; 
in vec3 colorPass[]; 
in float lifeTimePass[]; 
in float sizePass[]; 
in int typePass[]; 

// All that we send further

out vec3 positionOut; 
out vec3 velocityOut; 
out vec3 colorOut; 
out float lifeTimeOut; 
out float sizeOut; 
out int typeOut; 

uniform vec3 position; // Position where new particles are spawned
uniform vec3 gravityVector; // Gravity vector for particles - updates velocity of particles 
uniform vec3 direction; // direction of new particle normalized
uniform float angleOpen;  // angle in which the direction vary
uniform float minVelocity; 
uniform float maxVelocity;
uniform float slowdownFactor;

uniform vec3 startColor; 
uniform vec3 colorChange; 
uniform float size;  

uniform float lifeMin, lifeRange; // Life of new particle - from min to (min+range)
uniform float timePassed; // Time passed since last frame

uniform vec3 randomSeed; // Seed number for our random number function
vec3 localSeed; 

uniform int numToGenerate; // How many particles will be generated next time

// This function returns random number from zero to one
float randZeroOne() 
{ 
    uint n = floatBitsToUint(localSeed.y * 214013.0 + localSeed.x * 2531011.0 + localSeed.z * 141251.0); 
    n = n * (n * n * 15731u + 789221u); 
    n = (n >> 9u) | 0x3F800000u; 
  
    float res =  2.0 - uintBitsToFloat(n); 
    localSeed = vec3(localSeed.x + 147158.0 * res, localSeed.y*res  + 415161.0 * res, localSeed.z + 324154.0*res); 
    return res; 
} 

void main() 
{ 
  localSeed = randomSeed; 
   
  positionOut = positionPass[0]; 
  velocityOut = velocityPass[0]; 
  colorOut = colorPass[0]; 
  lifeTimeOut = lifeTimePass[0]-timePassed; 
  sizeOut = sizePass[0]; 
  typeOut = typePass[0]; 

  if(typePass[0] != 0) // existing particle
  { 
	positionOut += velocityOut*timePassed; 
	velocityOut = velocityOut - velocityOut*max(slowdownFactor*timePassed, 0) + gravityVector*timePassed; 
	colorOut += colorChange*timePassed;
  }

     
  if(typeOut == 0)  // generator particle
  { 
    EmitVertex(); 
    EndPrimitive(); 
     
    for(int i = 0; i < numToGenerate; i++) 
    { 
      positionOut = position; 
	  // calc random direction in a cone with angleOpen
	  float theta = cos(angleOpen) + (1 - cos(angleOpen))*randZeroOne();
	  float phi = randZeroOne() * 6.283185;
	  vec3 dir = vec3(sqrt(1-theta*theta)*cos(phi), 
				      sqrt(1-theta*theta)*sin(phi), 
					  theta);
	  if(direction.z == -1)
	  {
		 dir = dir * (-1);
	  }
	  else if(direction.z != 1)
	  {
		 vec3 axis = cross(dir, vec3(0, 0, 1));
		 float angle = acos(dot(dir, vec3(0, 0, 1)));
		 dir = cos(angle)*direction + 
		       sin(angle)*(cross(axis, direction)) + 
			   (1-cos(angle))*dot(axis, direction)*axis;
	  }
	  // calc velocity
	  float length =  minVelocity + (maxVelocity - minVelocity)*randZeroOne();
      velocityOut = dir * length; 
      colorOut = startColor; 
      lifeTimeOut = lifeMin+lifeRange*randZeroOne(); 
      sizeOut = size; 
      typeOut = 1; 
      EmitVertex(); 
      EndPrimitive(); 
    } 
  } 
  else if(lifeTimeOut > 0.0) 
  { 
      EmitVertex(); 
      EndPrimitive();  
  } 
}